#############################################################################
# pgceditor.tcl Copyright 2004 r0lZ
# The PGC Header editor stuff.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

namespace eval ::pgced {

namespace export edit_pgc
namespace import ::readdvd::get_int
namespace import ::writedvd::set_int


proc BCDtime {bcd showfps} {
	set pbtimetxt           [format  %i [expr  [get_int bcd 0 1] / 16]]
	set pbtimetxt $pbtimetxt[format  %i [expr  [get_int bcd 0 1] & 15]]
	set pbtimetxt $pbtimetxt:[format %i [expr  [get_int bcd 1 1] / 16]]
	set pbtimetxt $pbtimetxt[format  %i [expr  [get_int bcd 1 1] & 15]]
	set pbtimetxt $pbtimetxt:[format %i [expr  [get_int bcd 2 1] / 16]]
	set pbtimetxt $pbtimetxt[format  %i [expr  [get_int bcd 2 1] & 15]]
	set pbtimetxt $pbtimetxt.[format %i [expr ([get_int bcd 3 1]& 0x30)/16]]
	set pbtimetxt $pbtimetxt[format  %i [expr  [get_int bcd 3 1] & 15]]
	if $showfps {
		set fpsmode [expr [get_int bcd 3 1] & 0xC0]
		switch $fpsmode {
			192 {set pbtimetxt "$pbtimetxt (at 30 fps)"}
			64  {set pbtimetxt "$pbtimetxt (at 25 fps)"}
			default {set pbtimetxt "$pbtimetxt (illegal FPS)"}
		}
	}
	return $pbtimetxt
}

proc AddBCDtime {bcd1 bcd2} {
	set fpsmode [expr [get_int bcd1 3 1] & 0xC0]
	switch $fpsmode {
		192 {set fps 30}
		64  {set fps 25}
		default {set fps 0}
	}
	if {$fps == 0} {return "\000\000\000\000"}

	set f1 [expr ((([get_int bcd1 3 1]& 0x30)/16)*10) + ([get_int bcd1 3 1] & 15)]
	set s1 [expr  (([get_int bcd1 2 1] / 16) * 10) + ([get_int bcd1 2 1] & 15)]
	set m1 [expr  (([get_int bcd1 1 1] / 16) * 10) + ([get_int bcd1 1 1] & 15)]
	set h1 [expr  (([get_int bcd1 0 1] / 16) * 10) + ([get_int bcd1 0 1] & 15)]

	set f2 [expr ((([get_int bcd2 3 1]& 0x30)/16)*10) + ([get_int bcd2 3 1] & 15)]
	set s2 [expr  (([get_int bcd2 2 1] / 16) * 10) + ([get_int bcd2 2 1] & 15)]
	set m2 [expr  (([get_int bcd2 1 1] / 16) * 10) + ([get_int bcd2 1 1] & 15)]
	set h2 [expr  (([get_int bcd2 0 1] / 16) * 10) + ([get_int bcd2 0 1] & 15)]

	set f [expr $f1 + $f2]
	set s [expr $s1 + $s2]
	set m [expr $m1 + $m2]
	set h [expr $h1 + $h2]

	while {$f >= $fps} {
		set f [expr $f - $fps]
		incr s
	}
	while {$s >= 60} {
		set s [expr $s - 60]
		incr m
	}
	while {$m >= 60} {
		set m [expr $m - 60]
		incr h
	}

	set f "[expr ((($f / 10) * 16) + ($f % 10)) | $fpsmode]"
	set s "[expr (($s / 10) * 16) + ($s % 10)]"
	set m "[expr (($m / 10) * 16) + ($m % 10)]"
	set h "[expr (($h / 10) * 16) + ($h % 10)]"

	return "[format %c $h][format %c $m][format %c $s][format %c $f]"
}

proc edit_pgc {pgcheader numprograms programmap numcells cellpbtable cellpostable} {
	# get editables values from the header's binary data

	set ::pgced::pbtimetxt [BCDtime [string range $pgcheader 4 7] true]

	set ::pgced::pbmode    [get_int pgcheader 0xA2 1]
	if {$::pgced::pbmode == 0} {
		set pbmodetxt "sequential"
	} else {
		set pbprogs [expr $::pgced::pbmode & 0x7F]
		if {[expr $::pgced::pbmode & 0x80] == 0} {
			set pbmodetxt "random: $pbprogs programs"
		} else {
			set pbmodetxt "shuffle: $pbprogs programs"
		}
	}
	set ::pgced::pbmodetxt $pbmodetxt

	set ::pgced::puos      [get_int pgcheader 8 4]
	set ::pgced::nextpgcn  [get_int pgcheader 0x9C 2]
	set ::pgced::prevpgcn  [get_int pgcheader 0x9E 2]
	set ::pgced::grouppgcn [get_int pgcheader 0xA0 2]
	set ::pgced::stilltime [get_int pgcheader 0xA3 1]

	# build interface
	set w ".pgceditor"
	if {[winfo exists $w]} {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm iconname $w Dialog
	wm protocol $w WM_DELETE_WINDOW {set ::pgced::done 0}
	wm transient $w .
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	set stdcursor [$w cget -cursor]
	bind $w <Escape> {set ::pgced::done 0}

	frame $w.f1
	pack  $w.f1 -pady 2 -anchor w

	# misc PGC values
	frame $w.f1.fl
	pack  $w.f1.fl -fill x -padx 2 -pady 2 -side left
	frame $w.f1.fl.mf
	pack  $w.f1.fl.mf -padx 2
	label $w.f1.fl.mf.l0 -text "Playback Time "
	entry $w.f1.fl.mf.e0 -textvariable ::pgced::pbtimetxt -state readonly -justify center
	label $w.f1.fl.mf.l1 -text "Playback Mode "
	entry $w.f1.fl.mf.e1 -textvariable ::pgced::pbmodetxt -state readonly -justify center
	label $w.f1.fl.mf.l2 -text "Prohibited User Operations "
	button $w.f1.fl.mf.b2 -textvariable ::pgced::puos -pady 0 \
			-command {::pgced::puos_editor "Prohibited User Operations" .pgceditor; .pgceditor.f1.fl.mf.b2 configure -text $::pgced::puos}
	bind   $w.f1.fl.mf.b2 <3> {set ::pgced::puos 0; .pgceditor.f1.fl.mf.b2 configure -text "0"}
	label $w.f1.fl.mf.l3 -text "Previous PGCN "
	entry $w.f1.fl.mf.e3 -textvariable ::pgced::prevpgcn  -justify center \
			-validate key -validatecommand "::pgced::value_entry_validate 0 0xFFFF {%d} {%P}"
	label $w.f1.fl.mf.l4 -text "Next PGCN "
	entry $w.f1.fl.mf.e4 -textvariable ::pgced::nextpgcn  -justify center \
			-validate key -validatecommand "::pgced::value_entry_validate 0 0xFF {%d} {%P}"
	label $w.f1.fl.mf.l5 -text "Group PGCN "
	entry $w.f1.fl.mf.e5 -textvariable ::pgced::grouppgcn  -justify center \
			-validate key -validatecommand "::pgced::value_entry_validate 0 0xFF {%d} {%P}"
	label $w.f1.fl.mf.l6 -text "PGC Still Time (255=infinite) "
	entry $w.f1.fl.mf.e6 -textvariable ::pgced::stilltime  -justify center \
			-validate key -validatecommand "::pgced::value_entry_validate 0 0xFF {%d} {%P}"
	grid  $w.f1.fl.mf.l0 -row 0 -column 0 -sticky e -ipady 2
	grid  $w.f1.fl.mf.e0 -row 0 -column 1 -sticky ew
	grid  $w.f1.fl.mf.l1 -row 1 -column 0 -sticky e -ipady 2
	grid  $w.f1.fl.mf.e1 -row 1 -column 1 -sticky ew
	grid  $w.f1.fl.mf.l2 -row 2 -column 0 -sticky e -ipady 2
	grid  $w.f1.fl.mf.b2 -row 2 -column 1 -sticky ew
	grid  $w.f1.fl.mf.l3 -row 3 -column 0 -sticky e -ipady 2
	grid  $w.f1.fl.mf.e3 -row 3 -column 1 -sticky ew
	grid  $w.f1.fl.mf.l4 -row 4 -column 0 -sticky e -ipady 2
	grid  $w.f1.fl.mf.e4 -row 4 -column 1 -sticky ew
	grid  $w.f1.fl.mf.l5 -row 5 -column 0 -sticky e -ipady 2
	grid  $w.f1.fl.mf.e5 -row 5 -column 1 -sticky ew
	grid  $w.f1.fl.mf.l6 -row 6 -column 0 -sticky e -ipady 2
	grid  $w.f1.fl.mf.e6 -row 6 -column 1 -sticky ew

	# audio streams control
	frame $w.f1.mfa -borderwidth 2 -relief groove
	pack  $w.f1.mfa -padx 2 -ipadx 4 -ipady 2 -side left
	label $w.f1.mfa.tit -text "Audio\nStreams" -pady 0
	grid  $w.f1.mfa.tit -row 0 -column 0 -columnspan 2
	for {set a 1} {$a <= 8} {incr a} {
		label $w.f1.mfa.al$a -text "$a "
		entry $w.f1.mfa.ab$a -textvariable ::pgced::audiostream$a -state disabled \
				-disabledforeground black -relief raised -width 5 \
				-font $::config(fixed_width_font) -cursor $stdcursor
		bind  $w.f1.mfa.ab$a <1> "::pgced::set_audio ::pgced::audiostream$a"
		grid  $w.f1.mfa.al$a -row $a -column 0 -sticky e
		grid  $w.f1.mfa.ab$a -row $a -column 1 -sticky ew
	}

	# sub-picture streams control
	frame $w.f1.mfs -borderwidth 2 -relief groove
	pack  $w.f1.mfs -padx 2 -pady 1 -ipadx 4 -ipady 2 -side left
	label $w.f1.mfs.tit -text "Sub-picture Streams\n(for 4:3, wide, letterbox, pan&scan)" -pady 0
	grid  $w.f1.mfs.tit -row 0 -column 0 -columnspan 8
	for {set s 1} {$s <= 32} {incr s} {
		label $w.f1.mfs.al$s -text "$s "
		entry $w.f1.mfs.ab$s -textvariable ::pgced::subpicstream$s -state disabled \
				-disabledforeground black -relief raised -width 11 \
				-font $::config(fixed_width_font) -cursor $stdcursor
		bind  $w.f1.mfs.ab$s <1> "::pgced::set_subpic ::pgced::subpicstream$s"
		set row [expr (($s-1) % 8) + 1]
		set col [expr (($s-1) / 8) * 2]
		grid  $w.f1.mfs.al$s -row $row -column $col -sticky e
		grid  $w.f1.mfs.ab$s -row $row -column [expr $col+1] -sticky ew
	}

	# bottom frame
	frame  $w.f2
	pack   $w.f2 -expand 1 -fill both

	# color table
	frame $w.f2.mfc -borderwidth 2 -relief groove
	pack  $w.f2.mfc -side left -fill x -ipadx 2 -ipady 2 -padx 4
	label $w.f2.mfc.l0 -text "Sub-picture\nColor Lookup Table "
	pack  $w.f2.mfc.l0
	frame $w.f2.mfc.f1
	pack  $w.f2.mfc.f1
	set col 0
	set row 0
	for {set c 0} {$c < 16} {incr c} {
		button $w.f2.mfc.f1.b$c -width 2 -pady 0 -command "::pgced::set_color $c $w.f2.mfc.f1.b$c"
		grid   $w.f2.mfc.f1.b$c -row $row -column $col -ipadx 3 -ipady 3 -padx 1 -pady 1
		incr col
		if {$col == 4} {
			set col 0
			incr row
		}
	}
	frame $w.f2.mfc.pad -height 4
	pack  $w.f2.mfc.pad
	frame $w.f2.mfc.f2
	pack  $w.f2.mfc.f2
	button $w.f2.mfc.f2.bc -text "copy all"  -width 8 -command ::pgced::copy_colors  -pady 0
	button $w.f2.mfc.f2.bp -text "paste all" -width 8 -command ::pgced::paste_colors -pady 0
	pack   $w.f2.mfc.f2.bc -side left -padx 3
	pack   $w.f2.mfc.f2.bp -side left -padx 2


	### Programs and Cells table
	frame  $w.f2.fc -borderwidth 2 -relief groove
	pack   $w.f2.fc -side left -padx 2 -expand 1 -fill both -anchor w
	canvas $w.f2.fc.c -yscrollcommand "$w.f2.fc.vs set" -height 8 \
			-highlightcolor [$w.f2.fc cget -background]
	pack   $w.f2.fc.c -side left -expand 1 -fill both
	scrollbar $w.f2.fc.vs -command "$w.f2.fc.c yview" -orient vert
	pack   $w.f2.fc.vs -side left -fill y
	frame  $w.f2.fc.c.f	;# destroyed later

	### buttons
	frame  $w.fb
	pack   $w.fb -anchor e -padx 4 -pady 4
	button $w.fb.ok -text "OK" -width 10 -command {set ::pgced::done 1}
	pack   $w.fb.ok -side left -padx 3
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::pgced::done 0}
	pack   $w.fb.cancel -side left -padx 2

	wm title $w "PGC Editor - [$::widget(main_cmd_lab) cget -text]"
	$w.f1.fl.mf.b2 configure -text $::pgced::puos

	for {set c 0} {$c < 16} {incr c} {
		set yc_color [string range $pgcheader [expr 0xA4 + ($c*4)] [expr 0xA4 + ($c*4) + 3]]
		$w.f2.mfc.f1.b$c configure -background [yc2rgb $yc_color]
	}
	for {set a 1} {$a <= 8} {incr a} {
		binary scan [string range $pgcheader [expr 0x0A + ($a*2)] [expr 0x0A + ($a*2)]] c audio_status
		if {$audio_status == 0} {
			set ::pgced::audiostream$a "none"
		} else {
			set ::pgced::audiostream$a [expr $audio_status & 7]
		}
	}
	for {set s 1} {$s <= 32} {incr s} {
		binary scan [string range $pgcheader [expr 0x18 + ($s*4)] [expr 0x18 + ($s*4) + 3]] cccc s1 s2 s3 s4
		if {$s1 == 0} {
			set ::pgced::subpicstream$s "none"
		} else {
			set ::pgced::subpicstream$s [format "%2d %2d %2d %2d" [expr $s1 & 0x1F] $s2 $s3 $s4]
		}
	}


	### Programs and Cells table
	destroy $w.f2.fc.c.f
	frame $w.f2.fc.c.f
	pack  $w.f2.fc.c.f -anchor nw
	$w.f2.fc.c create window 0 0 -window $w.f2.fc.c.f -anchor nw

	if {$numcells == 0} {
		label $w.f2.fc.c.f.w_0_0 -text "No Cells"
		pack $w.f2.fc.c.f.w_0_0
	} else {
		label $w.f2.fc.c.f.w_0_0 -text "Prog."
		label $w.f2.fc.c.f.w_0_1 -text "Cell"
		label $w.f2.fc.c.f.w_0_2 -text "Type\nFlags"
		label $w.f2.fc.c.f.w_0_3 -text "Layer\nBreak"
		label $w.f2.fc.c.f.w_0_4 -text "Res-\ntricted"
		label $w.f2.fc.c.f.w_0_5 -text "Still\nTime"
		label $w.f2.fc.c.f.w_0_6 -text "Cell\nCmd"
		label $w.f2.fc.c.f.w_0_7 -text "Playback\nTime"
		label $w.f2.fc.c.f.w_0_8 -text "Start\nTime"
		label $w.f2.fc.c.f.w_0_9 -text "Entry\nVOBU\nsector"
		label $w.f2.fc.c.f.w_0_10 -text "First\nILVU\nEnd"
		label $w.f2.fc.c.f.w_0_11 -text "Last\nVOBU\nStart"
		label $w.f2.fc.c.f.w_0_12 -text "Last\nVOBU\nEnd"
		label $w.f2.fc.c.f.w_0_13 -text "VOB\nID"
		label $w.f2.fc.c.f.w_0_14 -text "Cell\nID"

		for {set c 1} {$c <= $numcells} {incr c} {
			set program_cell($c) ""
		}
		for {set p 1} {$p <= $numprograms} {incr p} {
			set program_cell([get_int programmap [expr $p-1] 1]) $p
		}
		set start_time "\000\000\000\000"
		for {set c 1} {$c <= $numcells} {incr c} {
			set cellidx1 [expr ($c - 1) * 24]
			set cellidx2 [expr ($c - 1) * 4]
			label $w.f2.fc.c.f.w_${c}_0 -text $program_cell($c)
			label $w.f2.fc.c.f.w_${c}_1 -text $c

			set type [get_int cellpbtable $cellidx1 1]
			button $w.f2.fc.c.f.w_${c}_2 -text $type -width 3 \
					-anchor e -pady 0 \
					-command "::pgced::cell_type_editor $w.f2.fc.c.f.w_${c}_2 ::pgced::var_${c}_3"
			if {[expr $type & 8] == 8} {
				set [subst ::pgced::var_${c}_3] 0
			} else {
				set [subst ::pgced::var_${c}_3] 1
			}
			checkbutton $w.f2.fc.c.f.w_${c}_3 -variable ::pgced::var_${c}_3 \
					-command "::pgced::set_layer_break ::pgced::var_${c}_3 $w.f2.fc.c.f.w_${c}_2"
			if {$c == 1} {$w.f2.fc.c.f.w_${c}_3 configure -state disabled}
			set  ::pgced::var_${c}_4 [expr [get_int cellpbtable [expr $cellidx1 + 1] 1] & 64]
			checkbutton $w.f2.fc.c.f.w_${c}_4 -variable ::pgced::var_${c}_4 \
					-variable ::pgced::var_${c}_4 -onvalue 64 -offvalue 0
			set  ::pgced::var_${c}_5 [get_int cellpbtable [expr $cellidx1 + 2] 1]
			entry $w.f2.fc.c.f.w_${c}_5 -textvariable ::pgced::var_${c}_5 -width 3 -justify right
			set  ::pgced::var_${c}_6 [get_int cellpbtable [expr $cellidx1 + 3] 1]
			entry $w.f2.fc.c.f.w_${c}_6 -textvariable ::pgced::var_${c}_6 -width 3 -justify right
			set cell_time [string range $cellpbtable [expr $cellidx1 + 4] [expr $cellidx1 + 7]]
			label $w.f2.fc.c.f.w_${c}_7 -text [BCDtime $cell_time false]
			set start_time [AddBCDtime $cell_time $start_time]
			label $w.f2.fc.c.f.w_${c}_8 -text [BCDtime $start_time false]
			label $w.f2.fc.c.f.w_${c}_9 -text [get_int cellpbtable [expr $cellidx1 +  8] 4]
			label $w.f2.fc.c.f.w_${c}_10 -text [get_int cellpbtable [expr $cellidx1 + 12] 4]
			label $w.f2.fc.c.f.w_${c}_11 -text [get_int cellpbtable [expr $cellidx1 + 16] 4]
			label $w.f2.fc.c.f.w_${c}_12 -text [get_int cellpbtable [expr $cellidx1 + 20] 4]
			label $w.f2.fc.c.f.w_${c}_13 -text [get_int cellpostable $cellidx2 2]
			label $w.f2.fc.c.f.w_${c}_14 -text [get_int cellpostable [expr $cellidx2 + 3] 1]
		}
		for {set c 0} {$c <= $numcells} {incr c} {
			for {set f 0} {$f <= 14} {incr f} {
				grid $w.f2.fc.c.f.w_${c}_$f -row $c -column $f
			}
		}
		for {set f 0} {$f <= 14} {incr f} {
			grid columnconf $w.f2.fc.c.f $f -minsize 28 -pad 2
		}
	}

	# inform scrollbar of the height of the canvas's internal window
	update
	$w.f2.fc.c configure -scrollregion "0 0 0 [winfo height $w.f2.fc.c.f]" \
			-width [winfo width $w.f2.fc.c.f]

	# window geometry
	#set maxheight [winfo height $w.f2.fc.c.f]
	#if {$maxheight > 200} {set maxheight 200}
	#$w.f2.fc.c configure -height $maxheight
	update
	set ww [winfo reqwidth $w]
	set wh [winfo reqheight $w]
	set wx [expr [winfo rootx .]+(([winfo width  .]-$ww)/2)]
	set wy [expr [winfo rooty .]+(([winfo height .]-$wh)/2)-25]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	$w.f2.fc.c configure -height 1
	wm geometry $w ${ww}x$wh+$wx+$wy
	wm resizable $w 1 1

	if {! [winfo ismapped $w]} {
		wm deiconify $w
	}
	grab set $w

	set ::pgced::done -1

	tkwait variable ::pgced::done
	grab release $w
	wm withdraw $w

	if $::pgced::done {
		if {$::pgced::puos      != ""} {set_int pgcheader 8    [expr $::pgced::puos] 4}
		if {$::pgced::nextpgcn  != ""} {set_int pgcheader 0x9C [expr $::pgced::nextpgcn] 2}
		if {$::pgced::prevpgcn  != ""} {set_int pgcheader 0x9E [expr $::pgced::prevpgcn] 2}
		if {$::pgced::grouppgcn != ""} {set_int pgcheader 0xA0 [expr $::pgced::grouppgcn] 2}
		if {$::pgced::stilltime != ""} {set_int pgcheader 0xA3 [expr $::pgced::stilltime] 1}
		for {set c 0} {$c < 16} {incr c} {
			set rgb_color [$w.f2.mfc.f1.b$c cget -background]
			set_int pgcheader [expr 0xA4 + (4*$c)] [rgb2yc $rgb_color] 4
		}
		for {set a 1} {$a <= 8} {incr a} {
			set varname ::pgced::audiostream$a
			if {[subst $$varname] == "none"} {
				set as 0
			} else {
				set as [expr [subst $$varname] | 0x80]
			}
			set_int pgcheader [expr 0x0A + ($a*2)] $as 1
		}
		for {set s 1} {$s <= 32} {incr s} {
			set varname ::pgced::subpicstream$s
			if {[subst $$varname] == "none"} {
				set ss 0
			} else {
				set sl [subst $$varname]
				set ss [expr 0x80000000 + \
						([lindex $sl 0] * 0x1000000) + \
						([lindex $sl 1] * 0x10000) + \
						([lindex $sl 2] * 0x100) + \
						[lindex $sl 3]]
			}
			set_int pgcheader [expr 0x18 + ($s*4)] $ss 4
		}

		for {set c 1} {$c <= $numcells} {incr c} {
			set cellidx1 [expr ($c - 1) * 24]
			set_int cellpbtable $cellidx1 [$w.f2.fc.c.f.w_${c}_2 cget -text] 1
			set tmp ::pgced::var_${c}_4
			set_int cellpbtable [expr $cellidx1 + 1] [subst $$tmp] 1
			set tmp ::pgced::var_${c}_5
			set_int cellpbtable [expr $cellidx1 + 2] [subst $$tmp] 1
			set tmp ::pgced::var_${c}_6
			set_int cellpbtable [expr $cellidx1 + 3] [subst $$tmp] 1
		}

		return [list $pgcheader $cellpbtable]
	} else {
		return ""
	}
}

proc value_entry_validate {min max action val} {
	if {$action != 0} {
		if {$val=="0"||$val=="0x"||$val=="0X"} {
			if {$min>0 && $val=="0"} bell
			return true
		}
		if {![string is integer $val]} {
			bell
			return false
		}
		if {$val < $min} {
			bell
			return false
		}
		if {$val > $max} {
			bell
			return false
		}
	}
	return true
}

proc rgb2yc {rgb_color} {
	if {[string range $rgb_color 0 0] == "#"} {
		set rgb_color [string range $rgb_color 1 end]
	}
	set r [expr 0x[string range $rgb_color 0 1]]
	set g [expr 0x[string range $rgb_color 2 3]]
	set b [expr 0x[string range $rgb_color 4 5]]

	set y  [expr round(( 0.257*$r) + (0.504*$g) + (0.098*$b) + 16)]
	set cr [expr round(( 0.439*$r) - (0.368*$g) - (0.071*$b) + 128)]
	set cb [expr round((-0.148*$r) - (0.291*$g) + (0.439*$b) + 128)]

	if {$y  < 0} {set y  0}
	if {$cr < 0} {set cr 0}
	if {$cb < 0} {set cb 0}
	if {$y  > 255} {set y  255}
	if {$cr > 255} {set cr 255}
	if {$cb > 255} {set cb 255}

	return [expr ($y*0x10000) + ($cr*0x100) + $cb]
}

proc yc2rgb {yc_color} {
	if {[string length $yc_color] == 4} {
		set yc_color [string range $yc_color 1 end]
	}
	binary scan $yc_color ccc y cr cb
	set y  [expr ($y  + 0x100) % 0x100]
	set cr [expr ($cr + 0x100) % 0x100]
	set cb [expr ($cb + 0x100) % 0x100]

	set r [expr round((1.164*($y-16)) + (1.596*($cr-128)))]
	set g [expr round((1.164*($y-16)) - (0.813*($cr-128)) - (0.392*($cb-128)))]
	set b [expr round((1.164*($y-16)) + (2.017*($cb-128)))]

	if {$r < 0} {set r 0}
	if {$g < 0} {set g 0}
	if {$b < 0} {set b 0}
	if {$r > 255} {set r 255}
	if {$g > 255} {set g 255}
	if {$b > 255} {set b 255}

	return [format "#%02X%02X%02X" $r $g $b]
}

proc set_color {colornum w} {
	set newcolor [tk_chooseColor -initialcolor [$w cget -background] -parent .pgceditor -title "Select color #$colornum"]
	if {$newcolor == ""} return
	$w configure -background $newcolor
}

proc copy_colors {} {
	set colors {}
	for {set c 0} {$c < 16} {incr c} {
		lappend colors [.pgceditor.f2.mfc.f1.b$c cget -background]
	}
	clipboard clear
	clipboard append $colors
}

proc paste_colors {} {
	set err [catch {set colors [clipboard get]}]
	if {$err || [llength $colors] != 16} {
		bell
		return
	}
	for {set c 0} {$c < 16} {incr c} {
		.pgceditor.f2.mfc.f1.b$c configure -background [lindex $colors $c]
	}
}

proc puos_editor {label parent} {
	# puos labels
	# note: GoUp (bit 4) is not used in PGC
	set puos_labels {
		{Time play or search}
		{PTT play or search}
		{Title play}
		{Stop}
		{GoUp Group (not used in PGCs)}
		{Time or PTT search}
		{TopPG or PrevPG search}
		{NextPG search}
		{Forward scan}
		{Backward scan}
		{Menu call - Title}
		{Menu call - Root}
		{Menu call - Subpicture}
		{Menu call - Audio}
		{Menu call - Angle}
		{Menu call - PTT}
		{Resume}
		{Button select or activate}
		{Still off}
		{Pause on}
		{Audio stream change}
		{Subpicture stream change}
		{Angle change}
		{Karaoke audio mix change}
		{Video presentation mode change}
	}

	set w ".puoseditor"
	if {![winfo exists $w]} {
		toplevel $w -class Dialog
		wm withdraw $w
		wm iconname $w Dialog
		wm protocol $w WM_DELETE_WINDOW {set ::pgced::puosdone 0}
		bind $w <Escape> "set ::pgced::puosdone 0"
		wm transient $w $parent
		if {$::tcl_platform(platform) == "windows"} {
			catch {wm attributes $w -toolwindow 1}
		}
		wm title $w "Prohibited User Operations"

		label $w.lab -text $label
		pack  $w.lab -fill x -padx 4

		frame $w.f1
		pack  $w.f1 -fill x -padx 4

		set puon 0
		foreach p $puos_labels {
			checkbutton $w.f1.puo$puon -text "$p ([expr round(pow(2,$puon))])" -anchor w -pady 0 -borderwidth 0 -variable ::pgced::puo_$puon
			pack $w.f1.puo$puon -fill x -pady 0
			incr puon
		}
		$w.f1.puo4 configure -state disabled	;# GoUp flag normally not used in PGC (?)
		frame  $w.fb
		pack   $w.fb -fill x -padx 2 -pady 4
		button $w.fb.clear -text "Clear all" -command {set ::pgced::tmppuos 0; ::pgced::set_puos_vars}
		pack   $w.fb.clear -side left -padx 5
		frame  $w.fb.pad
		pack   $w.fb.pad -side left -expand 1 -fill x
		button $w.fb.ok -text "OK" -width 6 -command {set ::pgced::puosdone 1}
		pack   $w.fb.ok -side left
		button $w.fb.cancel -text "Cancel" -width 6 -command {set ::pgced::puosdone 0}
		pack   $w.fb.cancel -side left -padx 5

		update
		set wx [expr [winfo rootx $parent]+(([winfo width  $parent]-[winfo reqwidth  $w])/2)]
		set wy [expr [winfo rooty $parent]+(([winfo height $parent]-[winfo reqheight $w])/2)-25]
		if {$wx < 10} {set wx 10}
		if {$wy < 10} {set wy 10}
		wm geometry $w +$wx+$wy
		wm resizable $w 0 0
	}
	set ::pgced::tmppuos $::pgced::puos
	set ::pgced::tmppuos [expr $::pgced::puos & 0x1FFFFEF]	;# disable GoUp flag, not used in PGC (?)
	::pgced::set_puos_vars
	$w.lab configure -text $label

	grab set $w
	wm deiconify $w

	set ::pgced::puosdone -1
	tkwait variable ::pgced::puosdone
	grab release $w
	if {$parent != "."} {
		grab set $parent
	}
	wm withdraw $w
	if $::pgced::puosdone {
		set ::pgced::puos 0
		for {set p 0} {$p <= 24} {incr p} {
			set varname ::pgced::puo_$p
			if [subst $$varname] {
				set ::pgced::puos [expr $::pgced::puos | round(pow(2,$p))]
			}
		}
	}
	return $::pgced::puos
}

proc set_puos_vars {} {
	for {set p 0} {$p <= 24} {incr p} {
		if {[expr $::pgced::tmppuos & round(pow(2,$p))] == 0} {
			set ::pgced::puo_$p 0
		} else {
			set ::pgced::puo_$p 1
		}
	}
}

proc set_audio {streamvar} {
	set w .audiosetup
	if {! [winfo exists $w]} {
		toplevel $w -class Dialog
		wm withdraw $w
		wm protocol $w WM_DELETE_WINDOW {set ::pgced::audiodone 0}
		bind $w <Escape> "set ::pgced::audiodone 0"
		wm title $w "Audio stream setup"
		wm resizable $w 0 0
		if {$::tcl_platform(platform) == "windows"} {
			catch {wm attributes $w -toolwindow 1}
		}
		label $w.l -text "Audio stream number:"
		pack  $w.l
		scale $w.s -from 0 -to 7 -variable "::pgced::audio_status"
		pack  $w.s -fill x -expand 1 -padx 4

		frame  $w.fb
		pack   $w.fb -padx 2 -pady 4 -anchor e
		button $w.fb.dis -text "Disable" -width 7 -command {set ::pgced::audiodone 2}
		pack   $w.fb.dis -side left -padx 4
		button $w.fb.ok -text "Set" -width 7 -command {set ::pgced::audiodone 1}
		pack   $w.fb.ok -side left
		button $w.fb.cancel -text "Cancel" -width 7 -command {set ::pgced::audiodone 0}
		pack   $w.fb.cancel -side left -padx 4

		update
		set wx [expr [winfo rootx .pgceditor]+(([winfo width  .pgceditor]-[winfo reqwidth  $w])/2)]
		set wy [expr [winfo rooty .pgceditor]+(([winfo height .pgceditor]-[winfo reqheight $w])/2)-25]
		if {$wx < 10} {set wx 10}
		if {$wy < 10} {set wy 10}
		wm geometry $w +$wx+$wy
		wm resizable $w 0 0
	}
	if {[subst $$streamvar] == "none"} {
		set ::pgced::audio_status 0
	} else {
		set ::pgced::audio_status [subst $$streamvar]
	}

	grab set $w
	wm deiconify $w

	set ::pgced::audiodone -1
	tkwait variable ::pgced::audiodone
	grab release $w
	grab set .pgceditor
	wm withdraw $w
	switch $::pgced::audiodone {
		1 {set $streamvar $::pgced::audio_status}
		2 {set $streamvar "none"}
	}
	return
}

proc set_subpic {streamvar} {
	set w .subpicsetup
	if {! [winfo exists $w]} {
		toplevel $w -class Dialog
		wm withdraw $w
		wm protocol $w WM_DELETE_WINDOW {set ::pgced::subpicdone 0}
		bind $w <Escape> "set ::pgced::subpicdone 0"
		wm title $w "Subpic streams setup"
		wm resizable $w 0 0
		if {$::tcl_platform(platform) == "windows"} {
			catch {wm attributes $w -toolwindow 1}
		}
		label $w.l -text "Sub-picture stream numbers:"
		pack  $w.l

		frame $w.f
		pack  $w.f -padx 4
		label $w.f.l1 -text "\n4:3"
		scale $w.f.s1 -from 0 -to 15 -variable "::pgced::subpic_1_status" -length 170
		label $w.f.l2 -text "\nwide"
		scale $w.f.s2 -from 0 -to 15 -variable "::pgced::subpic_2_status"
		label $w.f.l3 -text "\nletterbox"
		scale $w.f.s3 -from 0 -to 15 -variable "::pgced::subpic_3_status"
		label $w.f.l4 -text "\npan&scan"
		scale $w.f.s4 -from 0 -to 15 -variable "::pgced::subpic_4_status"
		grid  $w.f.l1 -row 0 -column 0 -sticky e
		grid  $w.f.s1 -row 0 -column 1 -sticky ew
		grid  $w.f.l2 -row 1 -column 0 -sticky e
		grid  $w.f.s2 -row 1 -column 1 -sticky ew
		grid  $w.f.l3 -row 2 -column 0 -sticky e
		grid  $w.f.s3 -row 2 -column 1 -sticky ew
		grid  $w.f.l4 -row 3 -column 0 -sticky e
		grid  $w.f.s4 -row 3 -column 1 -sticky ew

		frame  $w.fb
		pack   $w.fb -padx 2 -pady 4 -anchor e
		button $w.fb.dis -text "Disable" -width 7 -command {set ::pgced::subpicdone 2}
		pack   $w.fb.dis -side left -padx 4
		button $w.fb.ok -text "Set" -width 7 -command {set ::pgced::subpicdone 1}
		pack   $w.fb.ok -side left
		button $w.fb.cancel -text "Cancel" -width 7 -command {set ::pgced::subpicdone 0}
		pack   $w.fb.cancel -side left -padx 4

		update
		set wx [expr [winfo rootx .pgceditor]+(([winfo width  .pgceditor]-[winfo reqwidth  $w])/2)]
		set wy [expr [winfo rooty .pgceditor]+(([winfo height .pgceditor]-[winfo reqheight $w])/2)-25]
		if {$wx < 10} {set wx 10}
		if {$wy < 10} {set wy 10}
		wm geometry $w +$wx+$wy
		wm resizable $w 0 0
	}
	if {[subst $$streamvar] == "none"} {
		set ::pgced::subpic_1_status 0
		set ::pgced::subpic_2_status 0
		set ::pgced::subpic_3_status 0
		set ::pgced::subpic_4_status 0
	} else {
		set sublist [subst $$streamvar]
		set ::pgced::subpic_1_status [lindex $sublist 0]
		set ::pgced::subpic_2_status [lindex $sublist 1]
		set ::pgced::subpic_3_status [lindex $sublist 2]
		set ::pgced::subpic_4_status [lindex $sublist 3]
	}
	grab set $w
	wm deiconify $w

	set ::pgced::subpicdone -1
	tkwait variable ::pgced::subpicdone
	grab release $w
	grab set .pgceditor
	wm withdraw $w
	switch $::pgced::subpicdone {
		1 {set $streamvar [format "%2d %2d %2d %2d" $::pgced::subpic_1_status \
				$::pgced::subpic_2_status $::pgced::subpic_3_status $::pgced::subpic_4_status]}
		2 {set $streamvar "none"}
	}
	return
}

proc cell_type_editor {parentbutton layerbreakvar} {

	set w ".celltypeeditor"
	if {![winfo exists $w]} {
		toplevel $w -class Dialog
		wm withdraw $w
		wm iconname $w Dialog
		wm protocol $w WM_DELETE_WINDOW {set ::pgced::celltypedone 0}
		bind $w <Escape> "set ::pgced::celltypedone 0"
		wm transient $w .pgceditor
		if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
		wm title $w "Cell Type Flags"

		frame $w.f1
		pack  $w.f1 -fill x -padx 4

		checkbutton $w.f1.c0 -text "Seamless angle linked in DSI (1)" -anchor w \
				-borderwidth 0 -variable ::pgced::celltype_0 -offvalue 0 -onvalue 1
		pack $w.f1.c0 -fill x
		checkbutton $w.f1.c1 -text "SCR discontinuity (2)" -anchor w \
				-borderwidth 0 -variable ::pgced::celltype_1 -offvalue 0 -onvalue 2
		pack $w.f1.c1 -fill x
		checkbutton $w.f1.c2 -text "Interleaved (4)" -anchor w \
				-borderwidth 0 -variable ::pgced::celltype_2 -offvalue 0 -onvalue 4
		pack $w.f1.c2 -fill x
		checkbutton $w.f1.c3 -text "Seamless playback linked in PCI (8)" -anchor w \
				-borderwidth 0 -variable ::pgced::celltype_3 -offvalue 0 -onvalue 8
		pack $w.f1.c3 -fill x
		checkbutton $w.f1.c4 -text "Angle block (16)" -anchor w \
				-borderwidth 0 -variable ::pgced::celltype_4 -offvalue 0 -onvalue 16
		pack $w.f1.c4 -fill x

		frame $w.f1.ft -borderwidth 2 -relief groove
		pack  $w.f1.ft -fill x -pady 2 -padx 2

		label $w.f1.ft.lab -text "Angle type:" -anchor w
		pack  $w.f1.ft.lab -fill x
		radiobutton $w.f1.ft.r0 -variable ::pgced::celltype_67 -value 0 -anchor w \
				-text "normal (0)" -pady 0 -bd 0
		pack  $w.f1.ft.r0 -fill x
		radiobutton $w.f1.ft.r1 -variable ::pgced::celltype_67 -value 64 -anchor w \
				-text "first of angle block (64)" -pady 0 -bd 0
		pack  $w.f1.ft.r1 -fill x
		radiobutton $w.f1.ft.r2 -variable ::pgced::celltype_67 -value 128 -anchor w \
				-text "middle of angle block (128)" -pady 0 -bd 0
		pack  $w.f1.ft.r2 -fill x
		radiobutton $w.f1.ft.r3 -variable ::pgced::celltype_67 -value 192 -anchor w \
				-text "last of angle block (192)" -pady 0 -bd 0
		pack  $w.f1.ft.r3 -fill x

		frame  $w.fb
		pack   $w.fb -anchor e -padx 4 -pady 2
		button $w.fb.ok -text "OK" -width 6 -command {set ::pgced::celltypedone 1}
		pack   $w.fb.ok -side left
		button $w.fb.cancel -text "Cancel" -width 6 -command {set ::pgced::celltypedone 0}
		pack   $w.fb.cancel -side left -padx 5

		update
		set wx [expr [winfo rootx .pgceditor]+(([winfo width  .pgceditor]-[winfo reqwidth  $w])/2)]
		set wy [expr [winfo rooty .pgceditor]+(([winfo height .pgceditor]-[winfo reqheight $w])/2)-25]
		if {$wx < 10} {set wx 10}
		if {$wy < 10} {set wy 10}
		wm geometry $w +$wx+$wy
		wm resizable $w 0 0
	}

	set ct [$parentbutton cget -text]
	set ::pgced::celltype_0  [expr $ct & 1]
	set ::pgced::celltype_1  [expr $ct & 2]
	set ::pgced::celltype_2  [expr $ct & 4]
	set ::pgced::celltype_3  [expr $ct & 8]
	set ::pgced::celltype_4  [expr $ct & 16]
	set ::pgced::celltype_67 [expr $ct & 192]

	grab set $w
	wm deiconify $w

	set ::pgced::celltypedone -1
	tkwait variable ::pgced::celltypedone
	grab release $w
	grab set .pgceditor
	wm withdraw $w
	if $::pgced::celltypedone {
		set  r $::pgced::celltype_0
		incr r $::pgced::celltype_1
		incr r $::pgced::celltype_2
		incr r $::pgced::celltype_3
		incr r $::pgced::celltype_4
		incr r $::pgced::celltype_67
		$parentbutton configure -text $r
		if {$::pgced::celltype_3 == 8} {
			set $layerbreakvar 0
		} else {
			set $layerbreakvar 1
		}

		return $r
	} else {
		return ""
	}
}

proc set_layer_break {breakflag typebutton} {
	set typeval [$typebutton cget -text]
	if [subst $$breakflag] {
		set typeval [expr $typeval & 0xF7]
	} else {
		set typeval [expr $typeval | 8]
	}
	$typebutton configure -text $typeval
}

}	;# end of namespace ::pgced
