#############################################################################
# writedvd.tcl Copyright 2004 r0lZ
# write DVD IFOs to disk.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

## Global variables usage: see writedvd.tcl

namespace eval ::writedvd {

namespace export writedvd DVD_size set_int
namespace import ::readdvd::get_int

#############################################################################
## Procedure:  convert_hexcommand_to_binary
proc convert_hexcommand_to_binary {hexcmd} {
	set str ""
	for {set i 0} {$i < 23} {incr i 3} {
		set str [format "%s%c" $str "0x[string range $hexcmd $i [expr $i+1]]"]
	}
	return $str
}

#############################################################################
## Procedure:  d2c
## Convert a value to binary characters
proc d2c {val numbytes} {
	switch $numbytes {
		1 {return [binary format c $val]}
		2 {return [binary format S $val]}
		4 {return [binary format I $val]}
		8 {return [binary format W $val]}
	}
}

#############################################################################
## Procedure:  set_int
## Set an integer inside the binary data whose variable is passed in arguments
proc {set_int} {binary_upvar start_offset value length} {
	upvar $binary_upvar binarystring
	set binarystring [string replace $binarystring $start_offset \
			[expr $start_offset + $length - 1] [d2c $value $length]]
	return
}


#############################################################################
## Procedure:  sector_pad
## pad data to sector boundary.  return padded sector
proc sector_pad {data_upvar} {
	upvar $data_upvar data
	set remaider [expr [string length $data] % 0x800]
	if {$remaider > 0} {
		return $data[string repeat \000 [expr 0x800 - $remaider]]
	} else {
		return $data
	}
}

#############################################################################
## Procedure:  menu_vob_sectors_used
## return the number of sectors used by the menu VOB file of a specific VTS set (0 for VMG)
proc menu_vob_sectors_used {dvddir vtsnum} {
	if {$vtsnum == 0} {
		set vobname [file join $dvddir VIDEO_TS.VOB]
	} else {
		set vobname [file join $dvddir VTS_[format %02i $vtsnum]_0.VOB]
	}
	if [file exists $vobname] {
		return [expr [file size $vobname] / 0x800]
	} else {
		return 0
	}
}

#############################################################################
## Procedure:  vob_sectors_used
## return the number of sectors used by the VOB files of a specific VTS set
proc vob_sectors_used {dvddir vtsnum} {
	if {$vtsnum == 0} {
		return [menu_vob_sectors_used $dvddir 0]
	} else {
		set numsecs [menu_vob_sectors_used $dvddir $vtsnum]
		set vobbasename [file join $dvddir VTS_[format %02i $vtsnum]_]
		for {set n 1} {$n < 10} {incr n} {
			set vobname ${vobbasename}$n.VOB
			if [file exists $vobname] {
				incr numsecs [expr [file size $vobname] / 0x800]
			} else {
				break
			}
		}
		return $numsecs
	}
}

#############################################################################
## Procedure:  all_sectors_used
## return the number of sectors used by a specific VTS set (ie all VOBs + two
## times the IFO)
proc all_sectors_used {dvddir vtsnum} {
	if {$vtsnum == 0} {
		set ifosize [expr [file size [file join $dvddir "VIDEO_TS.IFO"]] / 0x800]
		return [expr [menu_vob_sectors_used $dvddir 0] + (2 * $ifosize)]
	} else {
		set ifosize [expr [file size \
				[file join $dvddir VTS_[format %02i $vtsnum]_0.IFO]] / 0x800]
		return [expr [vob_sectors_used $dvddir $vtsnum] + (2 * $ifosize)]
	}
}

##############################################################################
### Procedure:  commands_to_pgc
### replace the commands in the PGC raw datas by the hex commands gived in args.
### Correct the pointers within the PGC.  Return the modified PGC.
### The modified PGC will likely NOT have the same length as the original!
#proc commands_to_pgc {data header commandslist} {
#
#	### we need to cut the PGC into parts
#	for {set i 0} {$i < 4} {incr i} {
#		set start_offset($i) [get_int data [expr 0xE4 + (2 * $i)] 2]
#		if {$start_offset($i) == 0} {
#			set end_offset($i) 0
#		} else {
#			set end_offset($i) [string length $data]
#		}
#	}
#	set loweroffset [string length $data]
#	for {set i 0} {$i < 4} {incr i} {
#		if {$start_offset($i) > 0} {
#			for {set j 0} {$j < 4} {incr j} {
#				if {$start_offset($j) > $start_offset($i) \
#						&& $start_offset($j) < $end_offset($i)} {
#					set end_offset($i) $start_offset($j)
#				}
#			}
#			if {$start_offset($i) < $loweroffset} {
#				set loweroffset $start_offset($i)
#			}
#		}
#	}
#	#set header [string range $data 0 [expr $loweroffset-1]]
#	for {set i 0} {$i < 4} {incr i} {
#		if {$start_offset($i) == 0} {
#			set part($i) ""
#		} else {
#			set part($i) [string range $data $start_offset($i) [expr $end_offset($i)-1]]
#		}
#	}
#
#
#	### now, we have the old commands isolated.  Replace them with the edited ones.
#
#	# build the binary commands and concat them
#	set numpre  0
#	set numpost 0
#	set numcell 0
#	set numcmds 0
#	set cmds ""
#	foreach c $commandslist {
#		if {[string range $c 0 0] == "*"} {
#			set type [lindex $c 1]
#		} else {
#			set cmds "$cmds[convert_hexcommand_to_binary $c]"
#			incr numcmds
#			switch $type {
#				pre  {incr numpre}
#				post {incr numpost}
#				cell {incr numcell}
#			}
#		}
#	}
#	# add commands header
#	#if {$numcmds > 0} {
#	#}
#		set end_address [expr [string length $cmds]+7]
#		set cmds "[d2c $numpre 2][d2c $numpost 2][d2c $numcell 2][d2c $end_address 2]$cmds"
#	# replace the old commands by the new ones
#	set part(0) $cmds
#
#
#	### correct offsets in PGC header
#	set pos 0xE4
#	set offset [string length $header]
#	for {set i 0} {$i < 4} {incr i} {
#		if {$part($i) == ""} {
#			set_int header $pos 0 2
#		} else {
#			set_int header $pos $offset 2
#			incr offset [string length $part($i)]
#		}
#		incr pos 2
#	}
#
#
#	### concat all parts together
#	set pgc $header
#	for {set i 0} {$i < 4} {incr i} {
#		set pgc "$pgc$part($i)"
#	}
#	return $pgc
#}

#############################################################################
## Procedure:  rebuild_pgc
## rebuild the PGC raw datas by concanating all parts together.
## Correct the pointers within the PGC.  Return the PGC.
## The new PGC will likely NOT have the same length as the original!
proc rebuild_pgc {header commandslist numprograms programmap numcells cellpbtable cellpostable} {

	# build the binary commands and concat them together
	set numpre  0
	set numpost 0
	set numcell 0
	set numcmds 0
	set cmds ""
	foreach c $commandslist {
		if {[string range $c 0 0] == "*"} {
			set type [lindex $c 1]
		} else {
			set cmds "$cmds[convert_hexcommand_to_binary $c]"
			incr numcmds
			switch $type {
				pre  {incr numpre}
				post {incr numpost}
				cell {incr numcell}
			}
		}
	}
	# add commands header
	set end_address [expr [string length $cmds]+7]
	set cmds "[d2c $numpre 2][d2c $numpost 2][d2c $numcell 2][d2c $end_address 2]$cmds"


	### correct number of elements in PGC header
	set_int header 2 $numprograms 1
	set_int header 3 $numcells 1

	### correct start byte values in PGC header
	# start of commands
	set offset [string length $header]
	set_int header 0xE4 $offset 2
	incr offset [string length $cmds]
	# start of program map
	if {$numprograms == 0} {
		set_int header 0xE6 0 2
	} else {
		set_int header 0xE6 $offset 2
		incr offset [string length $programmap]
	}
	# start of cell playback table and cell position table
	if {$numcells == 0} {
		set_int header 0xE8 0 2
		set_int header 0xEA 0 2
	} else {
		set_int header 0xE8 $offset 2
		incr offset [string length $cellpbtable]
		set_int header 0xEA $offset 2
	}


	### concat all parts together
	return "$header$cmds$programmap$cellpbtable$cellpostable"
}

#############################################################################
## Procedure:  commands_to_PGCI_UT
## replace the PGCs in the PGCI_UT raw datas by the new commands and PGC parts
## of the PGCI_UT number gived in args.
## Return the modified PGCI_UT.
## The modified PGCI_UT will probably NOT have the same length as the original!
proc commands_to_PGCI_UT {data vtsnum} {
	if {$data == ""} {
		return ""
	}
	if {$::menus($vtsnum,numlus) == 0} {
		return $data
	}

	### get datas

	# get the PGCI_UT global header
	set pgci_ut_header [string range $data 0 7]
	set idx 8
	# get all the PGCI_UT's LUs global headers
	for {set lu 1} {$lu <= $::menus($vtsnum,numlus)} {incr lu} {
		set lu_definition($lu) [string range $data $idx [expr $idx+7]]
		set lu_offset($lu) [get_int lu_definition($lu) 4 4]
		incr idx 8
	}

	# get all LUs, and all PGCs in the LUs, and replace commands in the PGCs

	for {set lu 1} {$lu <= $::menus($vtsnum,numlus)} {incr lu} {
		set idx $lu_offset($lu)
		set lu_header($lu) [string range $data $idx [expr $idx+7]]
		set lu_end_byte($lu) [get_int lu_header($lu) 4 4]
		incr idx 8

		for {set pgc 1} {$pgc <= $::menus($vtsnum,$lu,numpgcs)} {incr pgc} {
			set lu_pgc_header($lu,$pgc) [string range $data $idx [expr $idx+7]]
			set lu_pgc_offset($lu,$pgc) [get_int lu_pgc_header($lu,$pgc) 4 4]
			incr idx 8
		}
#		for {set pgc 1} {$pgc < $::menus($vtsnum,$lu,numpgcs)} {incr pgc} {
#			set lu_pgc_last_byte($lu,$pgc) [expr $lu_pgc_offset($lu,[expr $pgc+1])-1]
#		}
#		set lu_pgc_last_byte($lu,$::menus($vtsnum,$lu,numpgcs)) $lu_end_byte($lu)
#
#		# process each PGCs
#		for {set pgc 1} {$pgc <= $::menus($vtsnum,$lu,numpgcs)} {incr pgc} {
#			set idx1 [expr $lu_offset($lu) + $lu_pgc_offset($lu,$pgc)]
#			set idx2 [expr $lu_offset($lu) + $lu_pgc_last_byte($lu,$pgc)]
#			# get PGC and replace commands
#			set newpgcdata($lu,$pgc) [commands_to_pgc [string range $data $idx1 $idx2] \
#					$::menus($vtsnum,$lu,$pgc,header) $::menus($vtsnum,$lu,$pgc)]
#		}
		# process each PGCs
		for {set pgc 1} {$pgc <= $::menus($vtsnum,$lu,numpgcs)} {incr pgc} {
			set newpgcdata($lu,$pgc) [rebuild_pgc \
					$::menus($vtsnum,$lu,$pgc,header) \
					$::menus($vtsnum,$lu,$pgc) \
					$::menus($vtsnum,$lu,$pgc,numprograms) \
					$::menus($vtsnum,$lu,$pgc,programmap) \
					$::menus($vtsnum,$lu,$pgc,numcells) \
					$::menus($vtsnum,$lu,$pgc,cellpbtable) \
					$::menus($vtsnum,$lu,$pgc,cellpostable)]
		}
	}

	### rebuild hierarchy with the new PGCs

	# rebuild all LUs
	for {set lu 1} {$lu <= $::menus($vtsnum,numlus)} {incr lu} {
		set ludata($lu) $lu_header($lu)
		for {set pgc 1} {$pgc <= $::menus($vtsnum,$lu,numpgcs)} {incr pgc} {
			set ludata($lu) $ludata($lu)$lu_pgc_header($lu,$pgc)
			set lu_pgc_offset_address($lu,$pgc) [expr [string length $ludata($lu)]-4]
		}
		for {set pgc 1} {$pgc <= $::menus($vtsnum,$lu,numpgcs)} {incr pgc} {
			set_int ludata($lu) $lu_pgc_offset_address($lu,$pgc) [string length $ludata($lu)] 4
			set ludata($lu) $ludata($lu)$newpgcdata($lu,$pgc)
		}
		set_int ludata($lu) 4 [expr [string length $ludata($lu)]-1] 4
	}

	# rebuild the whole table
	set outdata $pgci_ut_header
	for {set lu 1} {$lu <= $::menus($vtsnum,numlus)} {incr lu} {
		set outdata $outdata$lu_definition($lu)
		set lu_offset_address($lu) [expr [string length $outdata]-4]
	}
	for {set lu 1} {$lu <= $::menus($vtsnum,numlus)} {incr lu} {
		set_int outdata $lu_offset_address($lu) [string length $outdata] 4
		set outdata $outdata$ludata($lu)
	}
	set_int outdata 4 [expr [string length $outdata]-1] 4

	return $outdata
}


#############################################################################
## Procedure:  write_VMG_IFO
## recreate and save the VIDEO_TS.IFO file
proc write_VMG_IFO {dvddir} {

	### process FP-PGC commands

	# split the old VMGM_MAT into parts (header and FP-PGC)
	set headeroffset [get_int ::raw_sectors(0,VMGM_MAT) 0x84 4]
	set header [string range $::raw_sectors(0,VMGM_MAT) 0 [expr $headeroffset-1]]
	set fp_pgc [string range $::raw_sectors(0,VMGM_MAT) $headeroffset end]


	# insert commands in the FP-PGC
#	set fp_pgc [commands_to_pgc $fp_pgc $::pgcs(0,0,header) $::pgcs(0,0)]
	set fp_pgc [rebuild_pgc $::pgcs(0,0,header) $::pgcs(0,0) 0 "" 0 "" ""]

	# correct End byte of VMGI_MAT
	set_int header 0x80 [expr [string length $header]+[string length $fp_pgc]-1] 4
	# the First-Play PGC start byte (at 0x84) is not changed

	# recreate full VMGM_MAT
	set vmgm_mat $header$fp_pgc
	set vmgm_mat [sector_pad vmgm_mat]


	### insert menu commands in the VMGM_PGCI_UT
	set ::raw_sectors(0,VMGM_PGCI_UT) \
			[commands_to_PGCI_UT $::raw_sectors(0,VMGM_PGCI_UT) 0]


	### add other parts, and correct sector pointers in header
	set sector_table_element 0xC4
	set start_sector [expr [string length $vmgm_mat]/0x800]
	set data ""
	foreach p {VMG_PTT_SRPT VMGM_PGCI_UT VMG_PTL_MAIT \
			VMG_VTS_ATRT VMG_TXTDT_MG VMGM_C_ADT VMGM_VOBU_ADMAP} {
		if {$::raw_sectors(0,$p)!=""} {
			# VMGM_VOBU_ADMAP has the end_byte at position 0, and
			# VMG_TXTDT_MG at pos 16! All other tables at pos 4.
			switch $p {
				"VMG_TXTDT_MG" {set end_byte_pos 16}
				"VMGM_VOBU_ADMAP" {set end_byte_pos 0}
				default {set end_byte_pos 4}
			}
			set_int ::raw_sectors(0,$p) $end_byte_pos [expr [string length $::raw_sectors(0,$p)]-1] 4
			set paddedsector [sector_pad ::raw_sectors(0,$p)]
			set data $data$paddedsector
			set_int vmgm_mat $sector_table_element $start_sector 4
			incr start_sector [expr [string length $paddedsector]/0x800]
		} else {
			set_int vmgm_mat $sector_table_element 0 4
		}
		incr sector_table_element 4
	}
	# set end sector of IFO
	set_int vmgm_mat 0x1C [expr $start_sector-1] 4
	# set end sector of VMG Set (last sector of BUP)
	set menu_vob_sectors_used [menu_vob_sectors_used $dvddir 0]
	set_int vmgm_mat 0x0C [expr $start_sector + $menu_vob_sectors_used + $start_sector -1] 4
	# set start sector of menu VOB, if any
	if {$menu_vob_sectors_used > 0} {
		set_int vmgm_mat 0xC0 $start_sector 4
	} else {
		set_int vmgm_mat 0xC0 0 4
	}

	# concat all
	set full_file "$vmgm_mat$data"


	### write file
	set errnum [saveFile [file join $dvddir "VIDEO_TS.IFO"] binary $full_file]
	return $errnum
}

#############################################################################
## Procedure:  write_VTS_IFO
## recreate and save a VTS_xx_0.IFO file
proc write_VTS_IFO {dvddir vtsnum} {

	### insert TT commands in the VTS_PGCI

	set data $::raw_sectors($vtsnum,VTS_PGCI)
	set header_part [string range $data 0 [expr ($::pgcs($vtsnum,numpgcs)*8)+7]]
	set idx 8
	for {set pgc 1} {$pgc <= $::pgcs($vtsnum,numpgcs)} {incr pgc} {
		set vts_offset_address($pgc) [expr $idx+4]
		set vts_offset($pgc) [get_int header_part $vts_offset_address($pgc) 4]
		incr idx 8
	}
	for {set pgc 1} {$pgc < $::pgcs($vtsnum,numpgcs)} {incr pgc} {
		set vts_last_byte($pgc) [expr $vts_offset([expr $pgc+1])-1]
	}
	set vts_last_byte($::pgcs($vtsnum,numpgcs)) [get_int header_part 4 4]

#	# process each PGCs
#	for {set pgc 1} {$pgc <= $::pgcs($vtsnum,numpgcs)} {incr pgc} {
#		set newpgcdata($pgc) [commands_to_pgc \
#				[string range $data $vts_offset($pgc) $vts_last_byte($pgc)] \
#				$::pgcs($vtsnum,$pgc,header) $::pgcs($vtsnum,$pgc)]
#	}
	# process each PGCs
	for {set pgc 1} {$pgc <= $::pgcs($vtsnum,numpgcs)} {incr pgc} {
		set newpgcdata($pgc) [rebuild_pgc \
				$::pgcs($vtsnum,$pgc,header) \
				$::pgcs($vtsnum,$pgc) \
				$::pgcs($vtsnum,$pgc,numprograms) \
				$::pgcs($vtsnum,$pgc,programmap) \
				$::pgcs($vtsnum,$pgc,numcells) \
				$::pgcs($vtsnum,$pgc,cellpbtable) \
				$::pgcs($vtsnum,$pgc,cellpostable)]
	}

	# rebuild the VTS_PGCI
	set ::raw_sectors($vtsnum,VTS_PGCI) $header_part
	for {set pgc 1} {$pgc <= $::pgcs($vtsnum,numpgcs)} {incr pgc} {
		set_int ::raw_sectors($vtsnum,VTS_PGCI) $vts_offset_address($pgc) \
				[string length $::raw_sectors($vtsnum,VTS_PGCI)] 4
		set ::raw_sectors($vtsnum,VTS_PGCI) $::raw_sectors($vtsnum,VTS_PGCI)$newpgcdata($pgc)
	}

	### insert menu commands in the VTSM_PGCI_UT
	set ::raw_sectors($vtsnum,VTSM_PGCI_UT) \
			[commands_to_PGCI_UT $::raw_sectors($vtsnum,VTSM_PGCI_UT) $vtsnum]


	### add all parts together, and correct sector pointers in header
	set sector_table_element 0xC8
	set start_sector [expr [string length $::raw_sectors($vtsnum,VTSI_MAT)]/0x800]
	set data ""
	foreach p {VTS_PTT_SRPT VTS_PGCI VTSM_PGCI_UT VTS_TMAPTI \
			VTSM_C_ADT VTSM_VOBU_ADMAP VTS_C_ADT VTS_VOBU_ADMAP} {
		if {$::raw_sectors($vtsnum,$p)!=""} {
			# VTSM_VOBU_ADMAP and VTS_VOBU_ADMAP have the end_byte at position 0, not 4!
			switch $p {
				"VTSM_VOBU_ADMAP" -
				"VTS_VOBU_ADMAP" {set end_byte_pos 0}
				default {set end_byte_pos 4}
			}
			set_int ::raw_sectors($vtsnum,$p) $end_byte_pos \
					[expr [string length $::raw_sectors($vtsnum,$p)]-1] 4
			set paddedsector [sector_pad ::raw_sectors($vtsnum,$p)]
			set data $data$paddedsector

			set_int ::raw_sectors($vtsnum,VTSI_MAT) $sector_table_element $start_sector 4
			incr start_sector [expr [string length $paddedsector]/0x800]
		} else {
			set_int ::raw_sectors($vtsnum,VTSI_MAT) $sector_table_element 0 4
		}
		incr sector_table_element 4
	}
	# set end sector of IFO
	set_int ::raw_sectors($vtsnum,VTSI_MAT) 0x1C [expr $start_sector-1] 4
	# set end sector of VTS Set (last sector of BUP)
	set vob_sectors_used [vob_sectors_used $dvddir $vtsnum]
	set_int ::raw_sectors($vtsnum,VTSI_MAT) 0x0C [expr $start_sector + \
			$vob_sectors_used + $start_sector -1] 4
	# set start sector of menu VOB, if any
	set menu_vob_sectors_used [menu_vob_sectors_used $dvddir $vtsnum]
	if {$menu_vob_sectors_used > 0} {
		set_int ::raw_sectors($vtsnum,VTSI_MAT) 0xC0 $start_sector 4
	} else {
		set_int ::raw_sectors($vtsnum,VTSI_MAT) 0xC0 0 4
	}
	# set start sector of title VOBs
	if [file exists [file join $dvddir "VTS_[format %02i $vtsnum]_1.VOB"]] {
		set_int ::raw_sectors($vtsnum,VTSI_MAT) 0xC4 \
				[expr $start_sector + $menu_vob_sectors_used] 4
	} else {
		set_int ::raw_sectors($vtsnum,VTSI_MAT) 0xC4 0 4
	}

	# concat all
	set full_file "$::raw_sectors($vtsnum,VTSI_MAT)$data"

	### write file
	set filename [file join $dvddir "VTS_[format %02i $vtsnum]_0.IFO"]
	set errnum [saveFile $filename binary $full_file]
	return $errnum
}

#############################################################################
## Procedure:  correct_VTS_sectors
## correct VTS sectors in VIDEO_TS.IFO's TT_SRPT table.
## THIS PROC MUST BE USED AFTER ALL IFO FILES ARE SAVED TO DISK!
proc correct_VTS_sectors {dvddir} {
	# get sector usage for the VMG and all VTS
	set sector 0
	for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
		set vts_start_sector($vts) $sector
		incr sector [all_sectors_used $dvddir $vts]
	}

	# Read back VIDEO_TS.IFO file
	set filename [file join $dvddir "VIDEO_TS.IFO"]
	set data [readFile $filename binary]
	if {$data == ""} {return 1}

	# modify the TT_SRPT table
	set tt_srtt_offset [expr [get_int data 0xC4 4] * 0x800]
	set numtitles [get_int data $tt_srtt_offset 2]
	set vtserr 0
	for {set title 0} {$title < $numtitles} {incr title} {
		set title_offset [expr $tt_srtt_offset + 8 + ($title * 12)]
		set title_vtsn [get_int data [expr $title_offset+6] 1]
		if [info exists vts_start_sector($title_vtsn)] {
			set_int data [expr $title_offset+8] $vts_start_sector($title_vtsn) 4
		} else {
			# the Number of Titles Play Maps do not match the number of VTS!
			set txt "WARNING: There is something wrong in the VMG_PTT_SRPT table!"
			set txt "$txt\nTitle [expr $title+1] refer to VTS $title_vtsn, which doesn't exist."
			set txt "$txt\n\nDo you want to fix the value of the Number of Title Play Maps to $title"
			set txt "$txt\n(and the Length of Table value from [get_int data [expr $tt_srtt_offset+4] 4] to [expr $title_offset-$tt_srtt_offset-1]) ?"
			if {[expr $title+1] == $numtitles} {
				set txt "$txt\nWARNING: doing so will remove the entry for Title [expr $title+1].\n"
			} else {
				set txt "$txt\nWARNING: doing so will remove the entries for Titles [expr $title+1] to $numtitles."
				set txt "$txt\nSome of thoses Titles may refer to another (possibly existing) VTS."
			}
			set txt "$txt\n\nAnswer \"no\" to leave this entry unchanged and check the next one,"
			set txt "$txt\nor \"cancel\" to abort the Get VTS Sectors operation."
			set txt "$txt\nYou will need to fix the VMG_PTT_SRPT table yourself."
			set r [tk_messageBox -icon warning -title "PgcEdit: Get VTS Sectors" -parent . \
					-type yesnocancel -default no -message $txt]
			switch $r {
				yes {
					set_int data $tt_srtt_offset $title 2
					set_int data [expr $tt_srtt_offset+4] [expr $title_offset-$tt_srtt_offset-1] 4
					break
				}
				no {
					set vtserr 1
				}
				cancel {
					return -1999
				}
			}
		}
	}

	# Write back VIDEO_TS.IFO file
	set errnum [saveFile $filename binary $data]
	if {$vtserr && $errnum==0} {
		return -1999
	} else {
		return $errnum
	}
}
#############################################################################
## Procedure:  writedvd
## This is the public entry to write all IFOs to disk
proc writedvd {dvddir} {
	#set oldcursor [. cget -cursor]
	#. configure -cursor watch
	#update idletasks

	set errcode [write_VMG_IFO $dvddir]
	if {$errcode == 0} {
		for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
			set errcode [write_VTS_IFO $dvddir $vts]
			if {$errcode != 0} break
		}
	}

	### correct VTS sectors in VMG's TT_SRPT table
	if {$errcode == 0} {
		set errcode [correct_VTS_sectors $dvddir]
	}

	### save BUP as well
	if {$errcode == 0 || $errcode == -1999} {
		file copy -force -- "$dvddir/VIDEO_TS.IFO" "$dvddir/VIDEO_TS.BUP"
		for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
			file copy -force -- "$dvddir/VTS_[format %02i $vts]_0.IFO" \
					"$dvddir/VTS_[format %02i $vts]_0.BUP"
		}
	}
	#. configure -cursor $oldcursor
	return $errcode
}

#############################################################################
## Procedure:  correct_VTS_sectors
## correct VTS sectors in VIDEO_TS.IFO's TT_SRPT table.
## THIS PROC MUST BE USED AFTER ALL IFO FILES ARE SAVED TO DISK!
proc DVD_size {} {
	# get sector usage for the VMG and all VTS
	if {![info exists ::dvddir]} {
		bell
		return
	}

	set sectorsize 0
	for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
		incr sectorsize [all_sectors_used $::dvddir $vts]
	}
	set txt "Size: [expr round(($sectorsize / 512.0) * 100) / 100.0] MB. ([expr round(($sectorsize / 524288.0) * 1000) / 1000.0] GB.)"
	set txt "$txt\nNumber of sectors used: $sectorsize"
	set sectorsfree [expr 2297030 - $sectorsize]
	if {$sectorsfree >= 0} {
		set txt "$txt\nNumber of sectors free: [expr 2297030 - $sectorsize]"
	} else {
		set txt "$txt\nNumber of sectors to remove: [expr (2297030 - $sectorsize) * -1]"
	}
	set txt "$txt\nPercent full: [expr round(($sectorsize/2297030.0) * 10000) / 100.0]"
	tk_messageBox -icon info -title "PgcEdit: DVD Size" -parent . \
			-type ok -message $txt
}

} ;# end namespace ::writedvd
